#!/usr/bin/perl

=pod

=head1 NAME

    qotw21.pl

=head1 SYNTAX

    qotw21.pl task-file
    cat task-file | qotw21.pl

=head1 TASK

    You will write a program to perform scheduling.  As we all
    know, tasks sometimes take longer than expected.  Sometimes
    when this happens, the final deadline of the project is
    affected; sometimes it isn't.  For example, consider the four
    tasks A, B, C, and D.  B and C depend on A, which means that
    they cannot be started until A is finished. D depends on B and
    C, and cannot be started until both B and C are finished:

         .-> B .
       A        :-> D
         `-> C '

    Suppose we expect the four tasks to take the following times:

           A: 1 day
           B: 2 days
           C: 3 days
           D: 1 day

    Then we don't expect the project to be finished for at least 5
    days: one day to complete A and start C; 3 days to complete C
    and start D, and another day to finish D.  Any delay in any of
    the three tasks A, C, or D will cause the completion date to
    slip.  We say that A, C, and D are on the "critical path".
    But B is not on the critical path, because B can go on while C
    is going on, and can take up to one day longer than expected
    without delaying the start of D.

    You will write a program which will calculate critical paths.  The
    input to the program will be a file in the following format:

           A 1
           B 2 A
           C 3 A
           D 1 B C
           FINAL 0 D

    Each line represents one task.  The first field in each line
    is the name of the task.  The second field is the expected
    duration.  If there are any other fields, they are the names
    of other tasks which must be finished before this task can
    start.

    The program will find all the tasks on the critical path to
    the task named FINAL and will print them out, one per line.

    It may happen that the input specifies tasks that cannot possibly be
    completed.  For example:

           A 1 B
           B 1 A
           FINAL 0 A B

    Here A can't start until B is finished, but B can't start until A is
    finished.  In such cases, the program should diagnose an error.

=cut

#------------------------------------------------------------------------------
# pragmas and modules:
#------------------------------------------------------------------------------
use strict;
use warnings;
#use Data::Dumper; $Data::Dumper::Sortkeys = 1;

#------------------------------------------------------------------------------
# prototypes:
#------------------------------------------------------------------------------
sub allpre ($$);
sub allsuc ($$);
sub calcpathes ($$);


#------------------------------------------------------------------------------
# global variables:
#------------------------------------------------------------------------------
my %task;

#------------------------------------------------------------------------------
# read input file
#------------------------------------------------------------------------------
while (<>) {
    #print;
    chomp;
    my ($name, $time, @pre) = split /\s+/;
    if (exists $task{$name}) {
        die "Ambigous task definition '$_' in line $.: task '$name' ".
            "already defined!\n";
    }
    else {
        $task{$name} = { duration       => $time,
                         predecessor  => [ @pre ],
                       };
    }
}

#------------------------------------------------------------------------------
# test task integrity
#------------------------------------------------------------------------------

=head1 TESTS

    The following problems are tested by the program:

=over

=item Is there a task called 'FINAL'?

    If not, the program dies with the message
    C<No task 'FINAL' exists>.

=cut

# is there a task called 'FINAL' ?
die "No task 'FINAL' exists\n" unless exists $task{FINAL};

=item Are all given predecessor tasks known?

    If not the program dies with
    C<Task 'X' is not defined, but it is a
    preliminary task of task 'Y'>.

=cut

# are all preliminary tasks defined?
for my $t (keys %task) {
    #print "testing task '$t'\n";
    for my $p (@{ $task{$t}->{predecessor} }) {
        #print "testing preliminary task '$p'\n";
        die "Task '$p' is not defined, but it is a ",
            "preliminary task of task '$t'\n"
            unless exists $task{$p};
    }
}

=item Is a task a direct predecessor of itself?

    If this is the case, the program dies with
    C<Task 'X' is a preliminary task of itself>.

=cut

# is a task a preliminary of itself?
for my $t (keys %task) {
    for my $p (@{ $task{$t}->{predecessor} }) {
        die "Task '$p' is a preliminary task of itself\n"
            if $p eq $t;
    }
}


=item Is a task a indirect predecessor of itself?

    If this is the case, the program dies with
    C<preliminary circle with task 'X'>.

=item Are there any unused tasks for the task 'FINAL'?
    
    If this is the case, the program gives the warning
    C<unused tasks for 'FINAL' : 'X', 'Y', 'Z'>. 

=back

=cut


=head1 CALCULATIONS

    The following things are calculated by the program:

=over

=item The direct predecessors of all tasks.

    (These are given in the input directly.)

=item All predecessors of all tasks.

    Therefore we have to cycle through all the predecessors and
    collect them by doing so. (This is done in the function
    C<allpre>.)

=cut

#------------------------------------------------------------------------------
# calculate all predecessors:
# (...and test if there are preliminary circles)
#------------------------------------------------------------------------------
for my $t (keys %task) {
    my @pre_t = allpre($t, \%task);
    $task{$t}->{predecessorall} = [ sort @pre_t ];
}


#------------------------------------------------------------------------------
# warn if there are task unused for FINAL
#------------------------------------------------------------------------------
do {
    #my @prefinal = allpre('FINAL', \%task);
    #print "preliminyries of 'FINAL':\n", Dumper \@prefinal;
    my @prefinal = @{ $task{FINAL}->{predecessorall} };
    my %pf;
    $pf{$_} = 1 for ('FINAL', @prefinal);
    my @unused = grep { not exists $pf{$_} } keys %task;
    warn "unused tasks for 'FINAL' : '" .
         (join "', '", @unused) . "'\n"
         if @unused;
} if 0;


=item The level (or column) of each task.

    This means in which column the task can be printed. This is
    the first step for plotting the tasks. Tasks with no
    predecessors can be printed in column 0 (the left most column),
    all other tasks can be printed one column after the highest
    column of all its predecessors.

=cut

#------------------------------------------------------------------------------
# calculate levels of the tasks (0 = without predecessor...)
#------------------------------------------------------------------------------
{
#my $c=0;
    # 0 : all without predecessor
    # n : maximum of all predecessors + 1
    my @toview = keys %task;
    while (my $t = shift @toview) {
#++$c;
        #print "analyse '$t' ('", (join "', '", @toview), "')\n";
        my $lvl   = -1;
        my $ok    = 1;
        my @pre_t = @{ $task{$t}->{predecessorall} };
        #print "    '$t' has predecessorall ('", (join "', '", @{ $task{$t}->{predecessorall} }), "')\n";
        for my $pt (@pre_t) {
            #print "    pre '$pt' ";
            if (exists $task{$pt}->{level}) {
                $lvl = $task{$pt}->{level} if $task{$pt}->{level} > $lvl;
                #print "has level '", $task{$pt}->{level}, "'.\n";
            }
            else {
                $ok = 0;
                #print "has no level yet.\n";
                last;
            }
        }
        if ($ok) {
            $task{$t}->{level} = $lvl + 1;
            #print "task '$t' has level '", $task{$t}->{level}, "'.\n";
        }
        else {
            push @toview, $t; # view this one again later
            #print "task '$t' has no level yet and will be seen later.\n";
        }
#exit if $c > 9;
    }   
}


#------------------------------------------------------------------------------
# calculate successors:
#------------------------------------------------------------------------------

=item The direct successors of all tasks.

    Therefore we have to look at all tasks, if the given task
    'X' is a predeccessor of task 'Y', then 'Y' is a successor
    of task 'X'.

=cut

for my $t (keys %task) {
    #print "testing task '$t'\n";
    $task{$t}{successor} = [];
    for my $tt (keys %task) {
        #print "    against task '$tt'\n";
        if (grep { $_ eq $t } @{ $task{$tt}->{predecessor} }) {
            #print "    '$t' is preliminary task of '$tt'\n";
            push @{ $task{$t}{successor} }, $tt;
        }
    }
}

=item All successors of all tasks.

    Therefore we have to cycle through all the successors and
    collect them by doing so. (This is done in the function
    C<allsuc>.)

=cut

for my $t (keys %task) {
    my @suc_t = allsuc($t, \%task);
    $task{$t}->{successorall} = [ sort @suc_t ];
}


#print Dumper \%task;


#------------------------------------------------------------------------------
# calculate pathes to FINAL:
#------------------------------------------------------------------------------

=item The pathes to the task 'FINAL'.

    Therefore we have to cycle backwards through all the
    predecessors and split the possible pathes for each
    direct predeccessor we find.

=back

=cut

my @maxtasks;
my @pathes;
{
    @pathes = calcpathes('FINAL', \%task);
    my @duration;
    my $maxduration = 0;
    #print "final pathes are:\n", Dumper \@pathes;
    for my $path (@pathes) {
        my $duration = 0;
        for my $t (@$path) {
            $duration += $task{$t}->{duration};
        }
        #print "path '", (join "'-'", @$path), "' needs duration $duration\n";
        push @duration, $duration;
        $maxduration = $duration if $maxduration < $duration;
    }
    #print "maximal duration is $maxduration\n";
    my %maxtasks;
    for my $i (0 .. $#pathes) {
        if ($duration[$i] == $maxduration) {
            $maxtasks{$_} = 1 for @{ $pathes[$i] };
        }
    }
    @maxtasks = sort keys %maxtasks;
}
#print Dumper \@pathes;
#print "maximal tasks: '", (join "', '", @maxelem), "'\n";
print "$_\n" for @maxtasks, 'FINAL', '';


=head1 DATA STRUCTURES

    The program creates a HoH (hash of hashes), where the name
    of the tasks are the keys and the values are a hash with
    informations about this task. Here is an example for one
    entry of this HoH:

        'FINAL' => {
                     'duration'       => '0',
                     'level'          => 3,
                     'predecessor'    => [
                                           'D'
                                         ],
                     'predecessorall' => [
                                           'A',
                                           'B',
                                           'C',
                                           'D'
                                         ],
                     'successor'      => [],
                     'successorall'   => []
                   }

    The name of this HoH is C<%task>.

    The calculated pathes to the task 'FINAL' are stored in an
    AoA (array of arrays) called C<@pathes>. This is the structur
    of the AoA:

        (
          [
            'A',
            'B',
            'D'
          ],
          [
            'A',
            'C',
            'D'
          ]
        )
    
    The critical tasks are stored in a simple array called
    C<@maxtasks> and they are printed out one task by line.

=cut


sub allpre ($$) {
    #--------------------------------------------------------------------------
    # sub          : a l l p r e
    #--------------------------------------------------------------------------
    # author       : Christian Duehl
    # task         : Calculate all predecessors of the given task.
    # parameters   : 1) task
    #                2) Reference to hash with all tasks
    # return value : Array with all predecessors
    #--------------------------------------------------------------------------
    # 0.0.1 - 2004-08-07 - CD - created
    #--------------------------------------------------------------------------

    my ($t, $task) = @_;
    my %pre;
    my @toview = @{ $task->{$t}->{predecessor} };
    #print "allpre: testing task $t (", (join'-',@toview), ")\n";

    while (my $p = shift @toview) {
        #print "allpre-loop: testing task $p (", (join'-',@toview), ")\n";
        $pre{$p} = 1;
        for my $pp (@{ $task->{$p}->{predecessor} }) {
            #print "allpre-inner-loop: testing task $pp (", (join'-',@toview), ")\n";
            die "preliminary circle with task '$pp'\n" if exists $pre{$pp};
            push @toview, $pp;
        }
    }

    return keys %pre;
} # sub allpre


sub allsuc ($$) {
    #--------------------------------------------------------------------------
    # sub          : a l l s u c
    #--------------------------------------------------------------------------
    # author       : Christian Duehl
    # task         : Calculate all successors of the given task.
    # parameters   : 1) task
    #                2) Reference to hash with all tasks
    # return value : Array with all successors
    #--------------------------------------------------------------------------
    # 0.0.1 - 2004-08-07 - CD - created
    #--------------------------------------------------------------------------

    my ($t, $task) = @_;
    my %suc;
    my @toview = @{ $task->{$t}->{successor} };
    #print "allsuc: testing task $t (", (join'-',@toview), ")\n";

    while (my $p = shift @toview) {
        #print "allsuc-loop: testing task $p (", (join'-',@toview), ")\n";
        $suc{$p} = 1;
        for my $pp (@{ $task->{$p}->{successor} }) {
            #print "allsuc-inner-loop: testing task $pp (", (join'-',@toview), ")\n";
            die "successor circle with task '$pp'\n"
                if exists $suc{$pp}; # can not be
            push @toview, $pp;
        }
    }

    return keys %suc;
} # sub allsuc


sub calcpathes ($$) {
    #--------------------------------------------------------------------------
    # sub          : c a l c p a t h e s
    #--------------------------------------------------------------------------
    # author       : Christian Duehl
    # task         : Calculates all pathes to the given task.
    # parameters   : 1) task
    #                2) Reference to hash with all tasks
    # return value : Array with all pathes (a path is stored as an array).
    #--------------------------------------------------------------------------
    # 0.0.1 - 2004-08-07 - CD - created
    #--------------------------------------------------------------------------

    my ($t, $task)  = @_;
    my @pathes      = ();
    my @finalpathes = ();

    #print "calculating pathes for task '$t'\n";

    for my $s (@{ $task->{$t}->{predecessor} }) {
        #print "    successor '$s'\n";
        push @pathes, [ $s ];
    }

    while (my $p = shift @pathes) {
        #print "    working on path '", (join "'-'", @$p), "'\n";
        if (scalar @{ $task->{$$p[0]}->{predecessor} }) {
            push @pathes, [ $_, @$p ] for @{ $task->{$$p[0]}->{predecessor} };
        }
        else {
            push @finalpathes, [ @$p ];
        }
    }

    return @finalpathes;
} # sub calcpathes

=pod

=head1 EXAMPLES

=over

=item The example given in the task description

    crian@blitz:~/perl/qotw> cat qotw21_input1.txt
    A 1
    B 2 A
    C 3 A
    D 1 B C
    FINAL 0 D
    crian@blitz:~/perl/qotw> qotw21.pl qotw21_input1.txt
    A
    C
    D

=item Circle predecessors / successors ('A'-'B'-'A')

    crian@blitz:~/perl/qotw> cat qotw21_input2.txt
    A 1 B
    B 1 A
    FINAL 0 A B
    crian@blitz:~/perl/qotw> qotw21.pl qotw21_input2.txt
    preliminary circle with task 'B' at
    ./qotw21.pl line 432, <> line 3.


=item Multiple definition of one task ('A')

    crian@blitz:~/perl/qotw> cat qotw21_input3.txt
    A 1 B
    A 1 C
    B 1 C
    FINAL 0 C
    crian@blitz:~/perl/qotw> qotw21.pl qotw21_input3.txt
    Ambigous task definition 'A 1 C' in line 2: task 'A'
    already defined! at ./qotw21.pl line 102, <> line 2.

=item Undefined tasks ('C')

    crian@blitz:~/perl/qotw> cat qotw21_input4.txt
    A 1 B
    B 1 C
    FINAL 0 A B
    crian@blitz:~/perl/qotw> qotw21.pl qotw21_input4.txt
    Task 'C' is not defined, but it is a preliminary task
    of task 'B' at ./qotw21.pl line 145, <> line 3.

=item Longer circle predecessors / successors ('E'-'F'-'G'-'E')

    crian@blitz:~/perl/qotw> cat qotw21_input5.txt
    A 1
    B 2 A
    C 3 A
    D 1 B C
    E 1 G
    F 1 E
    G 1 F
    FINAL 0 D G
    crian@blitz:~/perl/qotw> qotw21.pl qotw21_input5.txt
    preliminary circle with task 'E' at
    ./qotw21.pl line 432, <> line 8.

=item Unused tasks ('E')

    crian@blitz:~/perl/qotw> cat qotw21_input6.txt
    A 1
    B 2 A
    C 3 A
    D 1 B C
    E 12 C
    FINAL 0 D
    crian@blitz:~/perl/qotw> qotw21.pl qotw21_input6.txt
    unused tasks for 'FINAL' : 'E' at
    ./qotw21.pl line 220, <> line 6.
    A
    C
    D

=back

=cut
